from abc import ABC
from enum import auto


class AlbTrafficEnum(ABC):
    """
    Enum class which tells what type of ingress and egress
    traffic should be applied to application loadbalancer.
    """
    # Allow traffic to/from whole internet.
    INTERNET = auto()
    # Allow traffic to/from within VPC only.
    VPC = auto()
    # Do not allow traffic.
    NONE = auto()



from typing import List, Optional
from aws_cdk import core, aws_elasticloadbalancingv2, aws_ec2, aws_certificatemanager
from aws_alb.alb_traffic_enum import AlbTrafficEnum
from aws_alb.loadbalancer_listeners import LoadBalancerListeners
from aws_alb.loadbalancer_sg import LoadBalancerSecurityGroup


class ApplicationLoadbalancer:
    """
    Manager class which creates a loadbalancer and its listeners and security groups.
    """
    def __init__(
            self,
            scope: core.Stack,
            prefix: str,
            vpc: aws_ec2.Vpc,
            loadbalancer_subnets: List[aws_ec2.Subnet],
            security_groups: Optional[List[aws_ec2.ISecurityGroup]] = None,
            inbound_traffic: Optional[AlbTrafficEnum] = None,
            outbound_traffic: Optional[AlbTrafficEnum] = None,
            certificate: Optional[aws_certificatemanager.CfnCertificate] = None
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation stack in which the resources should be added.
        :param prefix: A prefix for newly created resources.
        :param vpc: Virtual private cloud in which the security groups and a loadbalancer itself should be placed.
        :param loadbalancer_subnets: Subnets in which the loadbalancer can live.
        :param security_groups: Additional security groups for a loadbalancer.
        :param inbound_traffic: Inbound traffic for a loadbalancer.
        :param outbound_traffic: Outbound traffic for a loadbalancer.
        :param certificate: To enable HTTPS on a loadbalancer, provide a certificate.

        :return: No return.
        """
        self.__loadbalancer_default_security_group = LoadBalancerSecurityGroup(
            scope=scope,
            prefix=prefix,
            vpc=vpc,
            https_enabled=certificate is not None,
            inbound=inbound_traffic or AlbTrafficEnum.INTERNET,
            outbound=outbound_traffic or AlbTrafficEnum.INTERNET
        )

        security_groups = security_groups or []
        security_groups.append(self.__loadbalancer_default_security_group)

        self.__loadbalancer = aws_elasticloadbalancingv2.CfnLoadBalancer(
            scope,
            prefix + 'AppLoadBalancer',
            security_groups=[sg.security_group_id for sg in security_groups],
            subnets=[subnet.subnet_id for subnet in loadbalancer_subnets],
            type='application',
            scheme='internet-facing',
            name=prefix + 'AppLoadBalancer'
        )

        self.__loadbalancer_default_listeners = LoadBalancerListeners(
            scope=scope,
            prefix=prefix,
            application_loadbalancer=self.__loadbalancer,
            certificate=certificate
        )

    @property
    def loadbalancer(self) -> aws_elasticloadbalancingv2.CfnLoadBalancer:
        """
        Loadbalancer instance.
        More about loadbalancer:
        https://docs.aws.amazon.com/elasticloadbalancing/latest/application/introduction.html

        :return: Loadbalancer.
        """
        return self.__loadbalancer

    @property
    def default_listeners(self) -> LoadBalancerListeners:
        """
        Default listeners of the loadbalancer.
        More about listeners:
        https://docs.aws.amazon.com/elasticloadbalancing/latest/application/load-balancer-listeners.html

        :return: Listeners.
        """
        return self.__loadbalancer_default_listeners

    @property
    def default_security_group(self) -> LoadBalancerSecurityGroup:
        """
        Default security groups of the loadbalancer.
        More about security groups:
        https://docs.aws.amazon.com/vpc/latest/userguide/VPC_SecurityGroups.html

        :return: Security groups.
        """
        return self.__loadbalancer_default_security_group



from aws_cdk import aws_ec2
from aws_cdk.aws_ec2 import SecurityGroup, IVpc, Peer
from aws_cdk.core import Stack
from aws_alb.alb_traffic_enum import AlbTrafficEnum


class LoadBalancerSecurityGroup(SecurityGroup):
    """
    A class which manages a default security group for a loadbalancer.
    """
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            vpc: IVpc,
            https_enabled: bool,
            inbound: AlbTrafficEnum,
            outbound: AlbTrafficEnum
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation stack in which the resources should be added.
        :param prefix: A prefix for newly created resources.
        :param vpc: Virtual private cloud in which the security groups and a loadbalancer itself should be placed.
        :param https_enabled: Boolean telling whether to enable https prots.
        :param inbound: Inbound traffic peers for loadbalancer.
        :param outbound: Outbound traffic peers for loadbalancer.
        """
        super().__init__(
            scope=scope,
            id=prefix + 'LBSG',
            vpc=vpc,
            allow_all_outbound=False,
            description=f'A {prefix} load balancer security group.',
            security_group_name=prefix + 'AppLoadBalancerSG'
        )

        if inbound == AlbTrafficEnum.INTERNET:
            in_cidr = Peer.any_ipv4()
        elif inbound == AlbTrafficEnum.VPC:
            in_cidr = Peer.ipv4(vpc.vpc_cidr_block)
        else:
            in_cidr = None

        if outbound == AlbTrafficEnum.INTERNET:
            out_cidr = Peer.any_ipv4()
        elif outbound == AlbTrafficEnum.VPC:
            out_cidr = Peer.ipv4(vpc.vpc_cidr_block)
        else:
            out_cidr = None

        if in_cidr:
            self.add_ingress_rule(
                peer=in_cidr,
                connection=aws_ec2.Port(
                    protocol=aws_ec2.Protocol.TCP,
                    string_representation=prefix + 'LBSGIngress80',
                    from_port=80,
                    to_port=80
                )
            )

            if https_enabled:
                self.add_ingress_rule(
                    peer=in_cidr,
                    connection=aws_ec2.Port(
                        protocol=aws_ec2.Protocol.TCP,
                        string_representation=prefix + 'LBSGIngress443',
                        from_port=443,
                        to_port=443
                    )
                )

        if out_cidr:
            self.add_egress_rule(
                peer=out_cidr,
                connection=aws_ec2.Port(
                    protocol=aws_ec2.Protocol.TCP,
                    string_representation=prefix + 'LBSGEgress80',
                    from_port=80,
                    to_port=80
                )
            )

            if https_enabled:
                self.add_egress_rule(
                    peer=out_cidr,
                    connection=aws_ec2.Port(
                        protocol=aws_ec2.Protocol.TCP,
                        string_representation=prefix + 'LBSGEgress443',
                        from_port=443,
                        to_port=443
                    )
                )

        def cant_access(**kwargs):
            raise PermissionError('You can not access these methods.')

        # Make sure ingress and egress rules are not modifiable.
        self.add_egress_rule = cant_access
        self.add_ingress_rule = cant_access



from typing import Optional
from aws_cdk import core, aws_elasticloadbalancingv2, aws_certificatemanager


class LoadBalancerListeners:
    """
    A class which manages default listeners for a loadbalancer.
    """
    LISTENER_HTTP_PORT_1 = 80
    LISTENER_HTTPS_PORT_1 = 443
    LISTENER_HTTP_PORT_2 = 8000
    LISTENER_HTTPS_PORT_2 = 44300

    def __init__(
            self,
            scope: core.Stack,
            prefix: str,
            application_loadbalancer: aws_elasticloadbalancingv2.CfnLoadBalancer,
            certificate: Optional[aws_certificatemanager.CfnCertificate]
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation stack in which the resources should be added.
        :param prefix: A prefix for newly created resources.
        :param application_loadbalancer: Loadbalancer instance.
        :param certificate: A certificate for a loadbalancer to enable https.
        """
        self.__certificate = certificate

        self.__listener_http_1 = None
        self.__listener_https_1 = None
        self.__listener_http_2 = None
        self.__listener_https_2 = None

        self.__not_found_404_action = aws_elasticloadbalancingv2.CfnListener.ActionProperty(
            type='fixed-response',
            fixed_response_config=aws_elasticloadbalancingv2.CfnListener.FixedResponseConfigProperty(
                status_code='404',
                message_body='Not found.'
            )
        )

        self.__listener_http_1 = aws_elasticloadbalancingv2.CfnListener(
            scope, prefix + 'HttpListener1',
            port=self.LISTENER_HTTP_PORT_1,
            protocol='HTTP',
            load_balancer_arn=application_loadbalancer.ref,
            default_actions=[
                aws_elasticloadbalancingv2.CfnListener.ActionProperty(
                    type='redirect',
                    redirect_config=aws_elasticloadbalancingv2.CfnListener.RedirectConfigProperty(
                        status_code='HTTP_301',
                        host='#{host}',
                        path='/#{path}',
                        port=str(self.LISTENER_HTTPS_PORT_1),
                        query='#{query}',
                        protocol='HTTPS'
                    )
                ) if certificate else self.__not_found_404_action
            ]
        )

        if certificate:
            self.__listener_https_1 = aws_elasticloadbalancingv2.CfnListener(
                scope, prefix + 'HttpsListener1',
                certificates=[aws_elasticloadbalancingv2.CfnListener.CertificateProperty(
                    certificate_arn=certificate.ref
                )],
                port=self.LISTENER_HTTPS_PORT_1,
                protocol='HTTPS',
                load_balancer_arn=application_loadbalancer.ref,
                default_actions=[self.__not_found_404_action]
            )

        self.__listener_http_2 = aws_elasticloadbalancingv2.CfnListener(
            scope, prefix + 'HttpListener2',
            port=self.LISTENER_HTTP_PORT_2,
            protocol='HTTP',
            load_balancer_arn=application_loadbalancer.ref,
            default_actions=[
                aws_elasticloadbalancingv2.CfnListener.ActionProperty(
                    type='redirect',
                    redirect_config=aws_elasticloadbalancingv2.CfnListener.RedirectConfigProperty(
                        status_code='HTTP_301',
                        host='#{host}',
                        path='/#{path}',
                        port=str(self.LISTENER_HTTPS_PORT_2),
                        query='#{query}',
                        protocol='HTTPS'
                    )
                ) if certificate else self.__not_found_404_action
            ]
        )

        if certificate:
            self.__listener_https_2 = aws_elasticloadbalancingv2.CfnListener(
                scope, prefix + 'HttpsListener2',
                certificates=[aws_elasticloadbalancingv2.CfnListener.CertificateProperty(
                    certificate_arn=certificate.ref
                )],
                port=self.LISTENER_HTTPS_PORT_2,
                protocol='HTTPS',
                load_balancer_arn=application_loadbalancer.ref,
                default_actions=[self.__not_found_404_action]
            )

    @property
    def https_enabled(self) -> bool:
        """
        Tells whether https is enabled or not for a loadbalancer.

        :return: Boolean for whether https is enabled.
        """
        return self.__certificate is not None

    @property
    def production_http_listener(self) -> aws_elasticloadbalancingv2.CfnListener:
        """
        Listener instance for serving http traffic.

        :return: Http listener.
        """
        return self.__listener_http_1

    @property
    def production_https_listener(self) -> Optional[aws_elasticloadbalancingv2.CfnListener]:
        """
        Listener instance for serving https traffic.

        :return: Https listener.
        """
        return self.__listener_https_1

    @property
    def deployments_http_listener(self) -> aws_elasticloadbalancingv2.CfnListener:
        """
        Listener instance for serving test http traffic.

        :return: Test http listener.
        """
        return self.__listener_http_2

    @property
    def deployments_https_listener(self) -> Optional[aws_elasticloadbalancingv2.CfnListener]:
        """
        Listener instance for serving test https traffic.

        :return: Test https listener.
        """
        return self.__listener_https_2
